@@ -215,7 +215,7 @@ static int switch_list(void *handle, snd_switch_list_t *list, int interface, int
 	case 2:
 		return snd_ctl_pcm_playback_switch_list(handle, device, list);
 	case 3:
-		return snd_ctl_pcm_record_switch_list(handle, device, list);
+		return snd_ctl_pcm_capture_switch_list(handle, device, list);
 	case 4:
 		return snd_ctl_rawmidi_output_switch_list(handle, device, list);
 	case 5:
@@ -235,7 +235,7 @@ static int switch_read(void *handle, snd_switch_t *sw, int interface, int device
 	case 2:
 		return snd_ctl_pcm_playback_switch_read(handle, device, sw);
 	case 3:
-		return snd_ctl_pcm_record_switch_read(handle, device, sw);
+		return snd_ctl_pcm_capture_switch_read(handle, device, sw);
 	case 4:
 		return snd_ctl_rawmidi_output_switch_read(handle, device, sw);
 	case 5:
@@ -256,7 +256,7 @@ static int switch_write(void *handle, snd_switch_t *sw, int interface, int devic
 	case 2:
 		return snd_ctl_pcm_playback_switch_write(handle, device, sw);
 	case 3:
-		return snd_ctl_pcm_record_switch_write(handle, device, sw);
+		return snd_ctl_pcm_capture_switch_write(handle, device, sw);
 	case 4:
 		return snd_ctl_rawmidi_output_switch_write(handle, device, sw);
 	case 5:
@@ -300,7 +300,7 @@ static int determine_switches(void *handle, struct ctl_switch **csw, int interfa
 		bzero(&sw, sizeof(sw));
 		strncpy(sw.name, item->name, sizeof(sw.name));
 		if ((err = switch_read(handle, &sw, interface, device)) < 0) {
-			error("Cannot read switch for interface %i and device %i: %s", interface, device, snd_strerror(err));
+			error("Cannot read switch '%s' for interface %i and device %i: %s", sw.name, interface, device, snd_strerror(err));
 			free(list.pswitches);
 			return 1;
 		}
@@ -837,6 +837,8 @@ static void soundcard_setup_write_mixer_element(FILE * out, struct mixer_element
 			fprintf(out, "    ; Delay : Min %i Max %i\n", info->data.teffect1.min_delay, info->data.teffect1.max_delay);
 		if (info->data.teffect1.effect & SND_MIXER_EFF1_FEEDBACK)
 			fprintf(out, "    ; Feedback : Min %i Max %i\n", info->data.teffect1.min_feedback, info->data.teffect1.max_feedback);
+		if (info->data.teffect1.effect & SND_MIXER_EFF1_DEPTH_REAR)
+			fprintf(out, "    ; Depth rear : Min %i Max %i\n", info->data.teffect1.min_depth_rear, info->data.teffect1.max_depth_rear);
 		fprintf(out, "    element(\"%s\",%i,%i,_3D_Effect1(", mixer_element_name(&element->eid), element->eid.index, element->eid.type);
 		idx = 0;
 		if (element->data.teffect1.effect & SND_MIXER_EFF1_SW) 
@@ -857,6 +859,8 @@ static void soundcard_setup_write_mixer_element(FILE * out, struct mixer_element
 			fprintf(out, "%sdelay=%i", idx++ > 0 ? "," : "", element->data.teffect1.delay);
 		if (element->data.teffect1.effect & SND_MIXER_EFF1_FEEDBACK)
 			fprintf(out, "%sfeedback=%i", idx++ > 0 ? "," : "", element->data.teffect1.feedback);
+		if (element->data.teffect1.effect & SND_MIXER_EFF1_DEPTH_REAR)
+			fprintf(out, "%sdepth_rear=%i", idx++ > 0 ? "," : "", element->data.teffect1.depth_rear);
 		fprintf(out, "))\n");
 		break; 
 	default:
@@ -942,12 +946,12 @@ int soundcard_setup_write(const char *cfgfile, int cardno)
 				continue;
 			fprintf(out, "  pcm(\"%s\") {\n", pcm->info.name);
 			if (pcm->pswitches) {
-				fprintf(out, "    playback {");
+				fprintf(out, "    playback {\n");
 				soundcard_setup_write_switches(out, "      ", SND_INTERFACE_PCM, &pcm->pswitches);
 				fprintf(out, "    }\n");
 			}
 			if (pcm->rswitches) {
-				fprintf(out, "    record {");
+				fprintf(out, "    capture {\n");
 				soundcard_setup_write_switches(out, "      ", SND_INTERFACE_PCM, &pcm->rswitches);
 				fprintf(out, "    }\n");
 			}
@@ -958,12 +962,12 @@ int soundcard_setup_write(const char *cfgfile, int cardno)
 				continue;
 			fprintf(out, "  rawmidi(\"%s\") {\n", rawmidi->info.name);
 			if (rawmidi->oswitches) {
-				fprintf(out, "    output {");
+				fprintf(out, "    output {\n");
 				soundcard_setup_write_switches(out, "      ", SND_INTERFACE_RAWMIDI, &rawmidi->oswitches);
 				fprintf(out, "    }\n");
 			}
 			if (rawmidi->iswitches) {
-				fprintf(out, "    input {");
+				fprintf(out, "    input {\n");
 				soundcard_setup_write_switches(out, "      ", SND_INTERFACE_RAWMIDI, &rawmidi->iswitches);
 				fprintf(out, "    }\n");
 			}
